@@ -33,9 +33,12 @@ import (
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/util/clock"
 
+	"k8s.io/apimachinery/pkg/version"
+
 	velerov1api "github.com/vmware-tanzu/velero/pkg/apis/velero/v1"
 	pkgbackup "github.com/vmware-tanzu/velero/pkg/backup"
 	"github.com/vmware-tanzu/velero/pkg/builder"
+	"github.com/vmware-tanzu/velero/pkg/discovery"
 	"github.com/vmware-tanzu/velero/pkg/generated/clientset/versioned/fake"
 	informers "github.com/vmware-tanzu/velero/pkg/generated/informers/externalversions"
 	"github.com/vmware-tanzu/velero/pkg/metrics"
@@ -44,6 +47,7 @@ import (
 	"github.com/vmware-tanzu/velero/pkg/plugin/clientmgmt"
 	pluginmocks "github.com/vmware-tanzu/velero/pkg/plugin/mocks"
 	"github.com/vmware-tanzu/velero/pkg/plugin/velero"
+	velerotest "github.com/vmware-tanzu/velero/pkg/test"
 	"github.com/vmware-tanzu/velero/pkg/util/logging"
 )
 
@@ -168,8 +172,13 @@ func TestProcessBackupValidationFailures(t *testing.T) {
 				logger          = logging.DefaultLogger(logrus.DebugLevel, formatFlag)
 			)
 
+			apiServer := velerotest.NewAPIServer(t)
+			discoveryHelper, err := discovery.NewHelper(apiServer.DiscoveryClient, logger)
+			require.NoError(t, err)
+
 			c := &backupController{
 				genericController:      newGenericController("backup-test", logger),
+				discoveryHelper:        discoveryHelper,
 				client:                 clientset.VeleroV1(),
 				lister:                 sharedInformers.Velero().V1().Backups().Lister(),
 				backupLocationLister:   sharedInformers.Velero().V1().BackupStorageLocations().Lister(),
@@ -236,8 +245,13 @@ func TestBackupLocationLabel(t *testing.T) {
 				logger          = logging.DefaultLogger(logrus.DebugLevel, formatFlag)
 			)
 
+			apiServer := velerotest.NewAPIServer(t)
+			discoveryHelper, err := discovery.NewHelper(apiServer.DiscoveryClient, logger)
+			require.NoError(t, err)
+
 			c := &backupController{
 				genericController:      newGenericController("backup-test", logger),
+				discoveryHelper:        discoveryHelper,
 				client:                 clientset.VeleroV1(),
 				lister:                 sharedInformers.Velero().V1().Backups().Lister(),
 				backupLocationLister:   sharedInformers.Velero().V1().BackupStorageLocations().Lister(),
@@ -293,8 +307,14 @@ func TestDefaultBackupTTL(t *testing.T) {
 		)
 
 		t.Run(test.name, func(t *testing.T) {
+
+			apiServer := velerotest.NewAPIServer(t)
+			discoveryHelper, err := discovery.NewHelper(apiServer.DiscoveryClient, logger)
+			require.NoError(t, err)
+
 			c := &backupController{
 				genericController:      newGenericController("backup-test", logger),
+				discoveryHelper:        discoveryHelper,
 				backupLocationLister:   sharedInformers.Velero().V1().BackupStorageLocations().Lister(),
 				snapshotLocationLister: sharedInformers.Velero().V1().VolumeSnapshotLocations().Lister(),
 				defaultBackupTTL:       defaultBackupTTL.Duration,
@@ -339,6 +359,11 @@ func TestProcessBackupCompletions(t *testing.T) {
 				ObjectMeta: metav1.ObjectMeta{
 					Namespace: velerov1api.DefaultNamespace,
 					Name:      "backup-1",
+					Annotations: map[string]string{
+						"velero.io/source-cluster-k8s-major-version": "1",
+						"velero.io/source-cluster-k8s-minor-version": "16",
+						"velero.io/source-cluster-k8s-gitversion":    "v1.16.4",
+					},
 					Labels: map[string]string{
 						"velero.io/storage-location": "loc-1",
 					},
@@ -349,6 +374,7 @@ func TestProcessBackupCompletions(t *testing.T) {
 				Status: velerov1api.BackupStatus{
 					Phase:               velerov1api.BackupPhaseCompleted,
 					Version:             1,
+					FormatVersion:       "1.1.0",
 					StartTimestamp:      &timestamp,
 					CompletionTimestamp: &timestamp,
 					Expiration:          &timestamp,
@@ -367,6 +393,11 @@ func TestProcessBackupCompletions(t *testing.T) {
 				ObjectMeta: metav1.ObjectMeta{
 					Namespace: velerov1api.DefaultNamespace,
 					Name:      "backup-1",
+					Annotations: map[string]string{
+						"velero.io/source-cluster-k8s-major-version": "1",
+						"velero.io/source-cluster-k8s-minor-version": "16",
+						"velero.io/source-cluster-k8s-gitversion":    "v1.16.4",
+					},
 					Labels: map[string]string{
 						"velero.io/storage-location": "alt-loc",
 					},
@@ -377,6 +408,7 @@ func TestProcessBackupCompletions(t *testing.T) {
 				Status: velerov1api.BackupStatus{
 					Phase:               velerov1api.BackupPhaseCompleted,
 					Version:             1,
+					FormatVersion:       "1.1.0",
 					StartTimestamp:      &timestamp,
 					CompletionTimestamp: &timestamp,
 					Expiration:          &timestamp,
@@ -398,6 +430,11 @@ func TestProcessBackupCompletions(t *testing.T) {
 				ObjectMeta: metav1.ObjectMeta{
 					Namespace: velerov1api.DefaultNamespace,
 					Name:      "backup-1",
+					Annotations: map[string]string{
+						"velero.io/source-cluster-k8s-major-version": "1",
+						"velero.io/source-cluster-k8s-minor-version": "16",
+						"velero.io/source-cluster-k8s-gitversion":    "v1.16.4",
+					},
 					Labels: map[string]string{
 						"velero.io/storage-location": "read-write",
 					},
@@ -408,6 +445,7 @@ func TestProcessBackupCompletions(t *testing.T) {
 				Status: velerov1api.BackupStatus{
 					Phase:               velerov1api.BackupPhaseCompleted,
 					Version:             1,
+					FormatVersion:       "1.1.0",
 					StartTimestamp:      &timestamp,
 					CompletionTimestamp: &timestamp,
 					Expiration:          &timestamp,
@@ -426,6 +464,11 @@ func TestProcessBackupCompletions(t *testing.T) {
 				ObjectMeta: metav1.ObjectMeta{
 					Namespace: velerov1api.DefaultNamespace,
 					Name:      "backup-1",
+					Annotations: map[string]string{
+						"velero.io/source-cluster-k8s-major-version": "1",
+						"velero.io/source-cluster-k8s-minor-version": "16",
+						"velero.io/source-cluster-k8s-gitversion":    "v1.16.4",
+					},
 					Labels: map[string]string{
 						"velero.io/storage-location": "loc-1",
 					},
@@ -437,6 +480,7 @@ func TestProcessBackupCompletions(t *testing.T) {
 				Status: velerov1api.BackupStatus{
 					Phase:               velerov1api.BackupPhaseCompleted,
 					Version:             1,
+					FormatVersion:       "1.1.0",
 					Expiration:          &metav1.Time{now.Add(10 * time.Minute)},
 					StartTimestamp:      &timestamp,
 					CompletionTimestamp: &timestamp,
@@ -456,6 +500,11 @@ func TestProcessBackupCompletions(t *testing.T) {
 				ObjectMeta: metav1.ObjectMeta{
 					Namespace: velerov1api.DefaultNamespace,
 					Name:      "backup-1",
+					Annotations: map[string]string{
+						"velero.io/source-cluster-k8s-major-version": "1",
+						"velero.io/source-cluster-k8s-minor-version": "16",
+						"velero.io/source-cluster-k8s-gitversion":    "v1.16.4",
+					},
 					Labels: map[string]string{
 						"velero.io/storage-location": "loc-1",
 					},
@@ -466,6 +515,7 @@ func TestProcessBackupCompletions(t *testing.T) {
 				Status: velerov1api.BackupStatus{
 					Phase:               velerov1api.BackupPhaseCompleted,
 					Version:             1,
+					FormatVersion:       "1.1.0",
 					StartTimestamp:      &timestamp,
 					CompletionTimestamp: &timestamp,
 					Expiration:          &timestamp,
@@ -487,6 +537,11 @@ func TestProcessBackupCompletions(t *testing.T) {
 				ObjectMeta: metav1.ObjectMeta{
 					Namespace: velerov1api.DefaultNamespace,
 					Name:      "backup-1",
+					Annotations: map[string]string{
+						"velero.io/source-cluster-k8s-major-version": "1",
+						"velero.io/source-cluster-k8s-minor-version": "16",
+						"velero.io/source-cluster-k8s-gitversion":    "v1.16.4",
+					},
 					Labels: map[string]string{
 						"velero.io/storage-location": "loc-1",
 					},
@@ -497,6 +552,7 @@ func TestProcessBackupCompletions(t *testing.T) {
 				Status: velerov1api.BackupStatus{
 					Phase:               velerov1api.BackupPhaseFailed,
 					Version:             1,
+					FormatVersion:       "1.1.0",
 					StartTimestamp:      &timestamp,
 					CompletionTimestamp: &timestamp,
 					Expiration:          &timestamp,
@@ -516,6 +572,11 @@ func TestProcessBackupCompletions(t *testing.T) {
 				ObjectMeta: metav1.ObjectMeta{
 					Namespace: velerov1api.DefaultNamespace,
 					Name:      "backup-1",
+					Annotations: map[string]string{
+						"velero.io/source-cluster-k8s-major-version": "1",
+						"velero.io/source-cluster-k8s-minor-version": "16",
+						"velero.io/source-cluster-k8s-gitversion":    "v1.16.4",
+					},
 					Labels: map[string]string{
 						"velero.io/storage-location": "loc-1",
 					},
@@ -526,6 +587,7 @@ func TestProcessBackupCompletions(t *testing.T) {
 				Status: velerov1api.BackupStatus{
 					Phase:               velerov1api.BackupPhaseFailed,
 					Version:             1,
+					FormatVersion:       "1.1.0",
 					StartTimestamp:      &timestamp,
 					CompletionTimestamp: &timestamp,
 					Expiration:          &timestamp,
@@ -546,8 +608,26 @@ func TestProcessBackupCompletions(t *testing.T) {
 				backupper       = new(fakeBackupper)
 			)
 
+			apiServer := velerotest.NewAPIServer(t)
+
+			apiServer.DiscoveryClient.FakedServerVersion = &version.Info{
+				Major:        "1",
+				Minor:        "16",
+				GitVersion:   "v1.16.4",
+				GitCommit:    "FakeTest",
+				GitTreeState: "",
+				BuildDate:    "",
+				GoVersion:    "",
+				Compiler:     "",
+				Platform:     "",
+			}
+
+			discoveryHelper, err := discovery.NewHelper(apiServer.DiscoveryClient, logger)
+			require.NoError(t, err)
+
 			c := &backupController{
 				genericController:      newGenericController("backup-test", logger),
+				discoveryHelper:        discoveryHelper,
 				client:                 clientset.VeleroV1(),
 				lister:                 sharedInformers.Velero().V1().Backups().Lister(),
 				backupLocationLister:   sharedInformers.Velero().V1().BackupStorageLocations().Lister(),
@@ -584,7 +664,7 @@ func TestProcessBackupCompletions(t *testing.T) {
 			require.NoError(t, sharedInformers.Velero().V1().Backups().Informer().GetStore().Add(test.backup))
 
 			// add the default backup storage location to the clientset and the informer/lister store
-			_, err := clientset.VeleroV1().BackupStorageLocations(defaultBackupLocation.Namespace).Create(defaultBackupLocation)
+			_, err = clientset.VeleroV1().BackupStorageLocations(defaultBackupLocation.Namespace).Create(defaultBackupLocation)
 			require.NoError(t, err)
 
 			require.NoError(t, sharedInformers.Velero().V1().BackupStorageLocations().Informer().GetStore().Add(defaultBackupLocation))
@@ -694,6 +774,46 @@ func TestValidateAndGetSnapshotLocations(t *testing.T) {
 			expectedVolumeSnapshotLocationNames: []string{"aws-us-west-1", "some-name"},
 			expectedSuccess:                     true,
 		},
+		{
+			name:   "location name does not correspond to any existing location and snapshotvolume disabled; should return empty VSL and no error",
+			backup: defaultBackup().Phase(velerov1api.BackupPhaseNew).VolumeSnapshotLocations("random-name").SnapshotVolumes(false).Result(),
+			locations: []*velerov1api.VolumeSnapshotLocation{
+				builder.ForVolumeSnapshotLocation(velerov1api.DefaultNamespace, "aws-us-east-1").Provider("aws").Result(),
+				builder.ForVolumeSnapshotLocation(velerov1api.DefaultNamespace, "aws-us-west-1").Provider("aws").Result(),
+				builder.ForVolumeSnapshotLocation(velerov1api.DefaultNamespace, "some-name").Provider("fake-provider").Result(),
+			},
+			expectedVolumeSnapshotLocationNames: nil,
+			expectedSuccess:                     true,
+		},
+		{
+			name:   "duplicate locationName per provider and snapshotvolume disabled; should return empty VSL and no error",
+			backup: defaultBackup().Phase(velerov1api.BackupPhaseNew).VolumeSnapshotLocations("aws-us-west-1", "aws-us-west-1").SnapshotVolumes(false).Result(),
+			locations: []*velerov1api.VolumeSnapshotLocation{
+				builder.ForVolumeSnapshotLocation(velerov1api.DefaultNamespace, "aws-us-east-1").Provider("aws").Result(),
+				builder.ForVolumeSnapshotLocation(velerov1api.DefaultNamespace, "aws-us-west-1").Provider("aws").Result(),
+			},
+			expectedVolumeSnapshotLocationNames: nil,
+			expectedSuccess:                     true,
+		},
+		{
+			name:   "no location name for the provider exists, only one VSL created and snapshotvolume disabled; should return empty VSL and no error",
+			backup: defaultBackup().Phase(velerov1api.BackupPhaseNew).SnapshotVolumes(false).Result(),
+			locations: []*velerov1api.VolumeSnapshotLocation{
+				builder.ForVolumeSnapshotLocation(velerov1api.DefaultNamespace, "aws-us-east-1").Provider("aws").Result(),
+			},
+			expectedVolumeSnapshotLocationNames: nil,
+			expectedSuccess:                     true,
+		},
+		{
+			name:   "multiple location names for a provider, no default location and backup has no location defined, but snapshotvolume disabled, should return empty VSL and no error",
+			backup: defaultBackup().Phase(velerov1api.BackupPhaseNew).SnapshotVolumes(false).Result(),
+			locations: []*velerov1api.VolumeSnapshotLocation{
+				builder.ForVolumeSnapshotLocation(velerov1api.DefaultNamespace, "aws-us-west-1").Provider("aws").Result(),
+				builder.ForVolumeSnapshotLocation(velerov1api.DefaultNamespace, "aws-us-east-1").Provider("aws").Result(),
+			},
+			expectedVolumeSnapshotLocationNames: nil,
+			expectedSuccess:                     true,
+		},
 	}
 
 	for _, test := range tests {
